# (C) Copyright 2004-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

"""
Test that menu and toolbar actions are triggered.
"""

from functools import partial
import unittest

import pyface
from pyface.action.schema.api import ActionSchema, SGroup, SMenu, SMenuBar, SToolBar
from traits.has_traits import HasTraits
from traits.trait_types import Bool
from traitsui.menu import Action, ActionGroup, Menu, MenuBar, ToolBar
from traitsui.item import Item
from traitsui.view import View

from traitsui.tests._tools import (
    BaseTestMixin,
    create_ui,
    is_mac_os,
    is_null,
    requires_toolkit,
    reraise_exceptions,
    ToolkitName,
)

if is_null():
    raise unittest.SkipTest("Not supported using the null backend")


TestAction = Action(
    name="Test", action="test_clicked", tooltip="Click to test"
)


class DialogWithToolbar(HasTraits):
    """Test dialog with toolbar and menu."""

    action_successful = Bool(False)

    def test_clicked(self):
        self.action_successful = True

    menubar = MenuBar(Menu(ActionGroup(TestAction), name="&Test menu"))

    toolbar = ToolBar(ActionGroup(TestAction))

    traits_view = View(
        Item(
            label="Click the button on the toolbar or the menu item.\n"
            "The 'Action successful' element should turn to True."
        ),
        Item("action_successful", style="readonly"),
        menubar=menubar,
        toolbar=toolbar,
        buttons=[TestAction, "OK"],
    )


class DialogWithSchema(HasTraits):
    """Test dialog with toolbar and menu schemas."""

    action_successful = Bool(False)

    def test_clicked(self):
        self.action_successful = True

    menubar = SMenuBar(
        SMenu(
            SGroup(
                ActionSchema(
                    action_factory=lambda **traits: TestAction,
                ),
            ),
            name="&Test menu",
        ),
    )

    toolbar = SToolBar(
        SGroup(
            ActionSchema(
                action_factory=lambda **traits: TestAction,
            ),
        ),
    )

    traits_view = View(
        Item(
            label="Click the button on the toolbar or the menu item.\n"
            "The 'Action successful' element should turn to True."
        ),
        Item("action_successful", style="readonly"),
        menubar=menubar,
        toolbar=toolbar,
        buttons=[TestAction, "OK"],
    )


# ----- qt4 helper functions


def _qt_trigger_action(container_class, ui):
    toolbar = ui.control.findChild(container_class)
    action = toolbar.actions()[0]
    action.trigger()


def _qt_click_button(ui):
    from pyface.qt.QtGui import QDialogButtonBox

    bbox = ui.control.findChild(QDialogButtonBox)
    button = bbox.buttons()[1]
    button.click()


class TestActions(BaseTestMixin, unittest.TestCase):

    def _test_actions(self, trigger_action_func):
        """Template test for wx, qt4, menu, and toolbar testing."""
        # Behavior: when clicking on a menu or toolbar action,
        # the corresponding function should be executed

        # create dialog with toolbar adn menu
        dialog = DialogWithToolbar()
        with reraise_exceptions(), create_ui(dialog) as ui:

            # press toolbar or menu button
            trigger_action_func(ui)

            # verify that the action was triggered
            self.assertTrue(dialog.action_successful)

    # ----- Qt tests

    @requires_toolkit([ToolkitName.qt])
    def test_qt_toolbar_action(self):
        # Behavior: when clicking on a toolbar action, the corresponding
        # function should be executed

        # Bug: in the Qt4 backend, a
        # TypeError: perform() takes exactly 2 arguments (1 given) was raised
        # instead

        qt_trigger_toolbar_action = partial(
            _qt_trigger_action, pyface.ui.qt4.action.tool_bar_manager._ToolBar
        )

        self._test_actions(qt_trigger_toolbar_action)

    @requires_toolkit([ToolkitName.qt])
    def test_qt_menu_action(self):
        # Behavior: when clicking on a menu action, the corresponding function
        # should be executed

        # Bug: in the Qt4 backend, a
        # TypeError: perform() takes exactly 2 arguments (1 given) was raised
        # instead

        qt_trigger_menu_action = partial(
            _qt_trigger_action, pyface.ui.qt4.action.menu_manager._Menu
        )

        self._test_actions(qt_trigger_menu_action)

    @requires_toolkit([ToolkitName.qt])
    def test_qt_button_action(self):
        # Behavior: when clicking on a button action, the corresponding
        # function should be executed

        # Bug: in the Qt4 backend, a
        # TypeError: perform() takes exactly 2 arguments (1 given) was raised
        # instead

        self._test_actions(_qt_click_button)

    # ----- wx tests

    @unittest.skip(
        "Problem with triggering toolbar actions. Issue #428 and #1843."
    )
    @requires_toolkit([ToolkitName.wx])
    def test_wx_toolbar_action(self):
        # Behavior: when clicking on a toolbar action, the corresponding
        # function should be executed

        import wx

        def _wx_trigger_toolbar_action(ui):
            # long road to get at the Id of the toolbar button
            toolbar_control = ui.control.GetToolBar()
            toolbar_item = toolbar_control.tool_bar_manager.groups[0].items[0]
            toolbar_item_wrapper = toolbar_item._wrappers[0]
            control_id = toolbar_item_wrapper.control_id

            # build event that clicks the button
            click_event = wx.CommandEvent(
                wx.wxEVT_COMMAND_TOOL_CLICKED, control_id
            )

            # send the event to the toolbar
            toolbar_control.ProcessEvent(click_event)

        self._test_actions(_wx_trigger_toolbar_action)

    @requires_toolkit([ToolkitName.wx])
    def test_wx_button_action(self):
        # Behavior: when clicking on a button action, the corresponding
        # function should be executed

        import wx

        def _wx_trigger_button_action(ui):
            # long road to get at the Id of the toolbar button
            button_sizer = ui.control.GetSizer().GetChildren()[2].GetSizer()
            button = button_sizer.GetChildren()[0].GetWindow()

            control_id = button.GetId()

            # build event that clicks the button
            click_event = wx.CommandEvent(
                wx.wxEVT_COMMAND_BUTTON_CLICKED, control_id
            )

            # send the event to the toolbar
            ui.control.ProcessEvent(click_event)

        self._test_actions(_wx_trigger_button_action)

    # TODO: I couldn't find a way to press menu items programmatically for wx


class TestActionSchemas(BaseTestMixin, unittest.TestCase):
    def setUp(self):
        BaseTestMixin.setUp(self)

    def tearDown(self):
        BaseTestMixin.tearDown(self)

    def _test_actions(self, trigger_action_func):
        """Template test for wx, qt, menu, and toolbar testing."""
        # Behavior: when clicking on a menu or toolbar action,
        # the corresponding function should be executed

        # create dialog with toolbar adn menu
        dialog = DialogWithSchema()
        with reraise_exceptions(), create_ui(dialog) as ui:

            # press toolbar or menu button
            trigger_action_func(ui)

            # verify that the action was triggered
            self.assertTrue(dialog.action_successful)

    # ----- Qt tests

    @requires_toolkit([ToolkitName.qt])
    def test_qt_toolbar_action(self):
        # Behavior: when clicking on a toolbar action, the corresponding
        # function should be executed

        # Bug: in the Qt4 backend, a
        # TypeError: perform() takes exactly 2 arguments (1 given) was raised
        # instead

        qt_trigger_toolbar_action = partial(
            _qt_trigger_action, pyface.ui.qt4.action.tool_bar_manager._ToolBar
        )

        self._test_actions(qt_trigger_toolbar_action)

    @requires_toolkit([ToolkitName.qt])
    def test_qt_menu_action(self):
        # Behavior: when clicking on a menu action, the corresponding function
        # should be executed

        # Bug: in the Qt4 backend, a
        # TypeError: perform() takes exactly 2 arguments (1 given) was raised
        # instead

        qt_trigger_menu_action = partial(
            _qt_trigger_action, pyface.ui.qt4.action.menu_manager._Menu
        )

        self._test_actions(qt_trigger_menu_action)

    # ----- wx tests

    @unittest.skipIf(
        not is_mac_os,
        "Problem with triggering toolbar actions on Linux and Windows. Issue #428.",  # noqa: E501
    )
    @requires_toolkit([ToolkitName.wx])
    def test_wx_toolbar_action(self):
        # Behavior: when clicking on a toolbar action, the corresponding
        # function should be executed

        import wx

        def _wx_trigger_toolbar_action(ui):
            # long road to get at the Id of the toolbar button
            toolbar_control = ui.control.GetToolBar()
            toolbar_item = toolbar_control.tool_bar_manager.groups[0].items[0]
            toolbar_item_wrapper = toolbar_item._wrappers[0]
            control_id = toolbar_item_wrapper.control_id

            # build event that clicks the button
            click_event = wx.CommandEvent(
                wx.wxEVT_COMMAND_TOOL_CLICKED, control_id
            )

            # send the event to the toolbar
            toolbar_control.ProcessEvent(click_event)

        self._test_actions(_wx_trigger_toolbar_action)

    # TODO: I couldn't find a way to press menu items programmatically for wx


if __name__ == "__main__":
    # Execute from command line for manual testing
    vw = DialogWithToolbar()
    vw.configure_traits()



# (C) Copyright 2004-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" Define an editor that displays a string value as a title.
"""


from .editor import Editor

from pyface.api import HeadingText

# FIXME: TitleEditor (the editor factory for title editors) is a proxy class
# defined here just for backward compatibility. The class has been moved to
# traitsui.editors.title_editor.
from traitsui.editors.title_editor import TitleEditor


# -------------------------------------------------------------------------
#  '_TitleEditor' class:
# -------------------------------------------------------------------------


class _TitleEditor(Editor):

    def init(self, parent):
        """Finishes initializing the editor by creating the underlying toolkit
        widget.
        """
        self._control = HeadingText(parent=parent, create=False)
        self._control.create()
        self.control = self._control.control
        self.set_tooltip()

    def update_editor(self):
        """Updates the editor when the object trait changes external to the
        editor.
        """
        self._control.text = self.str_value

    def dispose(self):
        """Cleanly dispose of the editor.

        This ensures that the wrapped Pyface Widget is cleaned-up.
        """
        if self._control is not None:
            self._control.destroy()
        super().dispose()


SimpleEditor = _TitleEditor
CustomEditor = _TitleEditor
ReadonlyEditor = _TitleEditor
TextEditor = _TitleEditor



# (C) Copyright 2004-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" Defines the help interface for displaying the help associated with a
    Traits UI View object.
"""


from .toolkit import toolkit


def default_show_help(info, control):
    """Default handler for showing the help associated with a view."""
    toolkit().show_help(info.ui, control)


# The default handler for showing help
show_help = default_show_help


def on_help_call(new_show_help=None):
    """Sets a new global help provider function.

    The help provider function must have a signature of
    *function*(*info*, *control*), where *info* is a UIInfo object for the
    current view, and *control* is the UI control that invokes the function
    (typically, a **Help** button). It is provided in case the help provider
    needs to position the help window relative to the **Help** button.

    To retrieve the current help provider function, call this function with
    no arguments.

    Parameters
    ----------
    new_show_help : function
        The function to set as the new global help provider

    Returns
    -------
    previous : callable
        The previous global help provider function
    """
    global show_help

    result = show_help
    if new_show_help is not None:
        show_help = new_show_help
    return result



# (C) Copyright 2004-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

"""
Display an editable List of Strings.

The ListStrEditor uses a custom ListStrAdapter to define colors and tooltips.
"""

from traitsui.list_str_adapter import ListStrAdapter
from traits.api import HasTraits, List, Str
from traitsui.api import View, Item, ListStrEditor


# -- The adapter ----------------------------------------------------------

class HeadlinesListAdapter(ListStrAdapter):
    """ Custom adapter for string lists being edited in the ListStrEditor.

    - Style ALL CAPS titles in red.
    - Describe the capitalization in the tooltip.
    """

    def get_text_color(self, object, trait, row):
        try:
            row_data = getattr(object, trait)[row]
        except IndexError:
            return ""

        if row_data.isupper():
            return "red"
        return super().get_text_color(object, trait, row)

    def get_tooltip(self, object, trait, row):
        try:
            row_data = getattr(object, trait)[row]
        except IndexError:
            # if the ListStr is editable, IndexError indicates the empty next slots
            return ""

        if row_data.isupper():
            return "An uppercase headline"
        elif row_data.istitle():
            return "A headline in title case"
        elif row_data.islower():
            return "A fully lowercase headline"
        else:
            return self.tooltip


# -- The class providing the view -------------------------------------

class HeadlinesListDemo(HasTraits):

    headlines = List(Str)

    view = View(
        Item(
            "headlines",
            show_label=False,
            editor=ListStrEditor(
                title="List of headlines (hover for description)",
                adapter=HeadlinesListAdapter(
                    # the default tooltip for generic rows
                    tooltip="A headline"
                ),
                auto_add=True
            ),
            # (QT only) the tooltip shown in the area outside of rows
            # or when the row's tooltip is empty
            tooltip="List of headlines"
        ),
        title='List of headlines',
        width=320,
        height=370,
        resizable=True,
    )


# -- Set up the Demo ------------------------------------------------------

demo = HeadlinesListDemo(
    headlines=[
        "Shark bites man",
        "MAN BITES SHARK",
        "The London Stock Exchange Collapses",
        "she sells sea shells on the sea shore"
    ]
)

# Run the demo (in invoked from the command line):
if __name__ == "__main__":
    demo.configure_traits()



# (C) Copyright 2004-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

"""
Boolean editor (checkbox or text)

A Boolean (True/False) trait is displayed and edited as a checkbox, by default.

It can also be displayed as text 'True'/'False', either editable or read-only.

This example also shows how to listen for a change in a trait, and take
action when its value changes.

It also demonstrates how to add vertical space and a Label (plain text which is
not editable.)

Please refer to the `BooleanEditor API docs`_ for further information.

.. _BooleanEditor API docs: https://docs.enthought.com/traitsui/api/traitsui.editors.boolean_editor.html#traitsui.editors.boolean_editor.BooleanEditor
"""

from traits.api import HasTraits, Bool, Int
from traitsui.api import Item, Label, Group, View


class BooleanEditorDemo(HasTraits):
    """Defines the main BooleanEditor demo class."""

    # a boolean trait to view:
    my_boolean_trait = Bool()
    count_changes = Int(0)

    # When the trait's value changes, do something.
    # The listener method is named '_TraitName_changed', where
    # 'TraitName' is the name of the trait being monitored.
    def _my_boolean_trait_changed(self):
        self.count_changes += 1

    # Demo view
    traits_view = View(
        '10',  # vertical space
        Item('my_boolean_trait', style='simple', id='simple'),
        '10',  # vertical space
        # We put this label in its own group so that it will be left justified.
        # Otherwise it will line up with other edit fields (indented):
        Group(
            Label(
                'The same Boolean trait can also be displayed and edited as '
                'text (True/False):'
            )
        ),
        '10',  # vertical space
        Item(
            'my_boolean_trait',
            style='readonly',
            label='Read-only style',
            id='readonly',
        ),
        Item('my_boolean_trait', style='text', label='Text style', id='text'),
        '10',
        'count_changes',
        title='Boolean trait',
        buttons=['OK'],
        resizable=True,
    )


# Create the demo view (but do not yet display it):
demo = BooleanEditorDemo()

# Display and edit the demo (if invoked from the command line):
if __name__ == '__main__':
    demo.configure_traits()



# (C) Copyright 2004-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

"""Traits UI 'display only' video editor."""
from pyface.qt import is_qt5
from pyface.qt.QtCore import QPoint, Qt, QUrl, Signal
from pyface.qt.QtGui import QImage, QPainter, QPalette, QSizePolicy
from pyface.qt.QtMultimedia import (
    QAudio,
    QMediaPlayer,
    QVideoFrame,
)
from pyface.qt.QtMultimediaWidgets import QVideoWidget
from traits.api import (
    Any, Bool, Callable, Float, Instance, Range, Str, observe,
)
from traitsui.editors.video_editor import AspectRatio, MediaStatus, PlayerState

from .editor import Editor

#: Map from ApectRatio enum values to Qt aspect ratio behaviours.
aspect_ratio_map = {
    'ignore': Qt.AspectRatioMode.IgnoreAspectRatio,
    'keep': Qt.AspectRatioMode.KeepAspectRatio,
    'expand': Qt.AspectRatioMode.KeepAspectRatioByExpanding,
}

#: Map from PlayerState enum values to QMediaPlayer states.
if is_qt5:
    state_map = {
        'stopped': QMediaPlayer.State.StoppedState,
        'playing': QMediaPlayer.State.PlayingState,
        'paused': QMediaPlayer.State.PausedState,
    }
else:
    state_map = {
        'stopped': QMediaPlayer.PlaybackState.StoppedState,
        'playing': QMediaPlayer.PlaybackState.PlayingState,
        'paused': QMediaPlayer.PlaybackState.PausedState,
    }

#: Map from QMediaPlayer states to PlayerState enum values.
reversed_state_map = {value: key for key, value in state_map.items()}

#: Map from QMediaPlayer media status values to MediaStatus enum values.
media_status_map = {
    QMediaPlayer.MediaStatus.NoMedia: 'no_media',
    QMediaPlayer.MediaStatus.LoadingMedia: 'loading',
    QMediaPlayer.MediaStatus.LoadedMedia: 'loaded',
    QMediaPlayer.MediaStatus.StalledMedia: 'stalled',
    QMediaPlayer.MediaStatus.BufferingMedia: 'buffering',
    QMediaPlayer.MediaStatus.BufferedMedia: 'buffered',
    QMediaPlayer.MediaStatus.EndOfMedia: 'end',
    QMediaPlayer.MediaStatus.InvalidMedia: 'invalid',
}
if is_qt5:
    media_status_map[QMediaPlayer.MediaStatus.UnknownMediaStatus] = 'unknown'


if is_qt5:
    # These classes support dynamically modifying the video stream frames but
    # only work on Qt5
    from pyface.qt.QtMultimedia import QAbstractVideoSurface

    class ImageWidget(QVideoWidget):
        """Paints a QImage to the window body."""

        def __init__(self, parent=None, image_func=None):
            import numpy as np

            super().__init__(parent)
            self.image = QImage()
            self._np_image = np.zeros(shape=(0, 0, 4))
            self.painter = None
            self.resizeEvent(None)
            if image_func is None:
                # Don't bother with creating an ndarray version
                self.image_func = lambda image, bbox: image, self._np_image
            else:
                self.image_func = image_func

        def resizeEvent(self, event):
            s = self.size()
            self.width = s.width()
            self.height = s.height()

        def setImage(self, image):
            self.image, self._np_image = self.image_func(
                image, (self.width, self.height)
            )
            self.update()

        def paintEvent(self, event):
            super().paintEvent(event)
            if self.painter is None:
                self.painter = QPainter()
            self.painter.begin(self)
            if self.image:
                self.painter.drawImage(QPoint(0, 0), self.image)
            self.painter.end()

    class VideoSurface(QAbstractVideoSurface):

        frameAvailable = Signal(['QImage'])

        def __init__(self, widget=None):
            super().__init__()
            self.widget = widget

        def supportedPixelFormats(self, handleType):
            return [QVideoFrame.Format_RGB32]

        def present(self, frame):
            from pyface.qt.QtMultimedia import QAbstractVideoBuffer

            cloned_frame = QVideoFrame(frame)
            cloned_frame.map(QAbstractVideoBuffer.ReadOnly)
            image = QImage(
                cloned_frame.bits(),
                cloned_frame.width(),
                cloned_frame.height(),
                cloned_frame.bytesPerLine(),
                QVideoFrame.imageFormatFromPixelFormat(
                    cloned_frame.pixelFormat()
                ),
            )
            self.frameAvailable.emit(image)
            return True


class VideoEditor(Editor):
    """Traits UI 'display only' video editor.

    This editor uses the Qt QMultimedia machinery to display video streams
    to the screen.  Rather than being self-contained, the editor only concerns
    itself with displaying the video, and provides traits that control
    behaviour and provide internal state of the control during playback.
    """

    #: Does the drawing onto the image plane
    control = Instance(QVideoWidget)

    #: Handles the image pulling so the frames can be processed.  Qt5 only.
    surface = Any()

    #: The QMediaObject (Qt5) or QUrl (Qt6+) that holds the connection to the
    #: video stream.
    media_content = Any()

    #: The QMediaPlayer that controls playback of the video stream.
    media_player = Instance(QMediaPlayer)

    #: The aspect ratio of the video editor.
    aspect_ratio = AspectRatio()

    #: The current state of the player, synchronized to the trait named
    #: by factory.state.
    state = PlayerState()

    #: The current playback position of the player, synchronized to the trait
    #: named by factory.position.
    position = Float()

    #: The total duration of the current video, synchronized to the trait
    #: named by factory.duration.
    duration = Float()

    #: The media player playback status (loading, buffering, etc.),
    #: synchronized to the trait named by factory.media_status.
    media_status = MediaStatus()

    #: The percentage of the buffer currently filled, synchronized to the trait
    #: named by factory.buffer.
    buffer = Range(0, 100)

    #: A string holding the video error state, or "" if no error. Synchronized
    #: to the trait named by factory.video_error.
    video_error = Str()

    #: Whether the audio is muted or not, synchronized to the trait named by
    #: factory.muted.
    muted = Bool(False)

    #: The playback volume on a logarithmic scale (perceptually linear),
    #: synchronized to the trait named by factory.volume.
    volume = Range(0.0, 100.0)

    #: The playback rate.  Negative values rewind the video.
    #: Synchronized to the trait named by factory.playback_rate.
    playback_rate = Float(1.0)

    #: Function to apply to the image. Takes ref to new frame and a size tuple.
    #: Synchronized to the trait named by factory.image_func.
    image_func = Callable()

    #: The change in position required for an update to be emitted.
    #: Synchronized to the trait named by factory.notify_interval.
    #: This is only used on Qt5.
    notify_interval = Float(1.0)

    #: Qt6-specific QAudioOutput handler.
    _audio = Any()

    def update_to_regular(self):
        if self.surface is not None:
            self.surface.frameAvailable.disconnect(self.control.setImage)
            self.surface = None

        self.control = QVideoWidget()
        self.control.setSizePolicy(
            QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding
        )
        self.control.setBackgroundRole(QPalette.ColorRole.Window)
        self.media_player.setVideoOutput(self.control)

    def update_to_functional(self):
        self.control = ImageWidget(image_func=self.image_func)
        self.control.setSizePolicy(
            QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding
        )
        self.control.setBackgroundRole(QPalette.ColorRole.Window)

        self.surface = VideoSurface(widget=self.control)
        self.surface.frameAvailable.connect(self.control.setImage)

        self.media_player.setVideoOutput(self.surface)

    # ------------------------------------------------------------------------
    # Editor interface
    # ------------------------------------------------------------------------

    def init(self, parent):
        """Initialize the editor by creating the underlying toolkit widget.

        Parameters
        ----------
        parent : QWidget or None
            The parent widget for this widget.
        """
        self.control = QVideoWidget()
        self.control.setSizePolicy(
            QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding
        )
        self.control.setBackgroundRole(QPalette.ColorRole.Window)

        if is_qt5:
            self.media_player = QMediaPlayer(None, QMediaPlayer.VideoSurface)
        else:
            self.media_player = QMediaPlayer()
        self._set_video_url()
        self.media_player.setVideoOutput(self.control)
        if is_qt5:
            self.media_player.setMuted(self.muted)
        else:
            from pyface.qt.QtMultimedia import QAudioOutput

            self._audio = QAudioOutput()
            self._audio.setMuted(self.muted)
            self.media_player.setAudioOutput(self._audio)
        self._update_state()
        self._update_aspect_ratio()
        self._update_muted()
        self._update_volume()
        self._update_playback_rate()
        self._update_notify_interval()

        self._connect_signals()

        self.set_tooltip()

    def dispose(self):
        self._disconnect_signals()
        if self.media_player is not None:
            # Avoid a segfault if the media player is currently playing
            self.media_player.setVideoOutput(None)
            if not is_qt5:
                self.media_player.setAudioOutput(None)

        super().dispose()

    def update_editor(self):
        """Update the editor when the object trait changes externally."""
        self._set_video_url()

    # ------------------------------------------------------------------------
    # Private interface
    # ------------------------------------------------------------------------

    def _connect_signals(self):
        if self.media_player is not None:
            if is_qt5:
                self.media_player.stateChanged.connect(
                    self._state_changed_emitted
                )
                self.media_player.error.connect(self._error_emitted)
                self.media_player.bufferStatusChanged.connect(
                    self._buffer_status_changed_emitted
                )
                self.media_player.notifyIntervalChanged.connect(
                    self._notify_interval_changed_emitted
                )
            else:
                self.media_player.playbackStateChanged.connect(
                    self._state_changed_emitted
                )
                self.media_player.errorOccurred.connect(self._error_emitted)
                self.media_player.bufferProgressChanged.connect(
                    self._buffer_status_changed_emitted
                )
            self.media_player.positionChanged.connect(
                self._position_changed_emitted
            )
            self.media_player.durationChanged.connect(
                self._duration_changed_emitted
            )
            self.media_player.mediaStatusChanged.connect(
                self._media_status_changed_emitted
            )

    def _disconnect_signals(self):
        if self.media_player is not None:
            if is_qt5:
                self.media_player.stateChanged.disconnect(
                    self._state_changed_emitted
                )
                self.media_player.error.disconnect(self._error_emitted)
                self.media_player.bufferStatusChanged.disconnect(
                    self._buffer_status_changed_emitted
                )
                self.media_player.notifyIntervalChanged.disconnect(
                    self._notify_interval_changed_emitted
                )
            else:
                self.media_player.playbackStateChanged.disconnect(
                    self._state_changed_emitted
                )
                self.media_player.errorOccurred.disconnect(self._error_emitted)
                self.media_player.bufferProgressChanged.disconnect(
                    self._buffer_status_changed_emitted
                )
            self.media_player.positionChanged.disconnect(
                self._position_changed_emitted
            )
            self.media_player.durationChanged.disconnect(
                self._duration_changed_emitted
            )
            self.media_player.mediaStatusChanged.disconnect(
                self._media_status_changed_emitted
            )

    def _set_video_url(self):
        qurl = QUrl.fromUserInput(self.value)
        if is_qt5:
            from pyface.qt.QtMultimedia import QMediaContent
            if qurl.isValid():
                self.media_content = QMediaContent(qurl)
            else:
                self.media_content = QMediaContent(None)
        else:
            self.media_content = qurl
        self.control.updateGeometry()

    # Signal handlers -------------------------------------------------------

    def _state_changed_emitted(self, state):
        self.state = reversed_state_map[state]

    def _position_changed_emitted(self, position):
        # Avoid telling Qt about the new position in `_position_changed`
        with self.updating_value():
            self.position = position / 1000.0

    def _duration_changed_emitted(self, duration):
        self.duration = duration / 1000.0

    def _error_emitted(self, error):
        if error != QMediaPlayer.Error.NoError:
            self.video_error = self.media_player.errorString()
        else:
            self.video_error = ''

    def _media_status_changed_emitted(self, error):
        self.media_status = media_status_map[self.media_player.mediaStatus()]

    def _buffer_status_changed_emitted(self, error):
        if is_qt5:
            self.buffer = self.media_player.bufferStatus()
        else:
            self.buffer = int(self.media_player.bufferProgress() * 100)

    def _notify_interval_changed_emitted(self, interval):
        self.notify_interval = interval / 1000.0

    # Trait change handlers -------------------------------------------------

    @observe('aspect_ratio')
    def _aspect_ratio_observer(self, event):
        if self.control is not None:
            self._update_aspect_ratio()

    @observe('image_func')
    def _image_func_observer(self, event):
        if self.image_func is None:
            self.update_to_regular()
        elif not is_qt5:
            raise ValueError("image_func is not supported on Qt6")
        else:
            self.update_to_functional()

    @observe('media_content')
    def _media_content_observer(self, event):
        self.video_error = ''
        if self.media_player is not None:
            if is_qt5:
                self.media_player.setMedia(self.media_content)
            else:
                self.media_player.setSource(self.media_content)

    @observe('muted')
    def _muted_observer(self, event):
        if self.media_player is not None:
            self._update_muted()

    @observe('playback_rate')
    def _playback_rate_observer(self, event):
        if self.media_player is not None:
            self._update_playback_rate()

    @observe('position')
    def _position_observer(self, event):
        if self.media_player is not None and not self.updating:
            # position is given in ms
            self.media_player.setPosition(int(self.position * 1000))

    @observe('state')
    def _state_observer(self, event):
        if self.media_player is not None:
            self._update_state()

    @observe('volume')
    def _volume_observer(self, event):
        if self.media_player is not None:
            self._update_volume()

    @observe('notify_interval')
    def _notify_interval_observer(self, event):
        if self.media_player is not None:
            self._update_notify_interval()

    # MediaPlayer management ------------------------------------------------

    def _update_aspect_ratio(self):
        self.control.setAspectRatioMode(aspect_ratio_map[self.aspect_ratio])

    def _update_muted(self):
        if is_qt5:
            self.media_player.setMuted(self.muted)
        else:
            self._audio.setMuted(self.muted)

    def _update_playback_rate(self):
        self.media_player.setPlaybackRate(self.playback_rate)

    def _update_state(self):
        if self.state == 'stopped':
            self.media_player.stop()
            self.control.repaint()
        elif self.state == 'playing':
            # XXX forcing a resize so video is scaled correctly on MacOS
            s = self.control.size()
            w = s.width()
            h = s.height()
            self.media_player.play()
            self.control.resize(w + 1, h + 1)
            self.control.resize(w, h)
        elif self.state == 'paused':
            self.media_player.pause()

    def _update_volume(self):
        linear_volume = QAudio.convertVolume(
            self.volume / 100.0,
            QAudio.VolumeScale.LogarithmicVolumeScale,
            QAudio.VolumeScale.LinearVolumeScale,
        )
        if is_qt5:
            self.media_player.setVolume(int(linear_volume * 100))
        else:
            self._audio.setVolume(linear_volume)

    def _update_notify_interval(self):
        # only used on Qt5
        if is_qt5:
            # interval is given in ms
            interval = int(self.notify_interval * 1000)
            self.media_player.setNotifyInterval(interval)



# (C) Copyright 2008-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

# ------------------------------------------------------------------------------
# Copyright (c) 2007, Riverbank Computing Limited
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD license.
# However, when used with the GPL version of PyQt the additional terms described
# in the PyQt GPL exception also apply

#
# Author: Riverbank Computing Limited
# ------------------------------------------------------------------------------

""" Defines the various list editors for the PyQt user interface toolkit.
"""


from pyface.qt import QtCore, QtGui, is_pyside

from pyface.api import ImageResource

from traits.api import (
    Any,
    Bool,
    Callable,
    Dict,
    Instance,
    List,
    Str,
    TraitError,
)
from traits.trait_base import user_name_for, xgetattr

from traitsui.editors.list_editor import ListItemProxy

from .editor import Editor
from .helper import IconButton
from .menu import MakeMenu


class SimpleEditor(Editor):
    """Simple style of editor for lists, which displays a list box with only
    one item visible at a time. A icon next to the list box displays a menu of
    operations on the list.

    """

    # -------------------------------------------------------------------------
    #  Trait definitions:
    # -------------------------------------------------------------------------

    #: The kind of editor to create for each list item
    kind = Str()

    #: Is the list of items being edited mutable?
    mutable = Bool(True)

    #: Is the editor scrollable?
    scrollable = Bool(True)

    #: Signal mapper allowing to identify which icon button requested a context
    #: menu
    mapper = Instance(QtCore.QSignalMapper)

    buttons = List([])

    _list_pane = Instance(QtGui.QWidget)

    # -------------------------------------------------------------------------
    #  Class constants:
    # -------------------------------------------------------------------------

    #: Whether the list is displayed in a single row
    single_row = True

    # -------------------------------------------------------------------------
    #  Normal list item menu:
    # -------------------------------------------------------------------------

    #: Menu for modifying the list
    list_menu = """
       Add &Before     [_menu_before]: self.add_before()
       Add &After      [_menu_after]:  self.add_after()
       ---
       &Delete         [_menu_delete]: self.delete_item()
       ---
       Move &Up        [_menu_up]:     self.move_up()
       Move &Down      [_menu_down]:   self.move_down()
       Move to &Top    [_menu_top]:    self.move_top()
       Move to &Bottom [_menu_bottom]: self.move_bottom()
    """

    # -------------------------------------------------------------------------
    #  Empty list item menu:
    # -------------------------------------------------------------------------

    empty_list_menu = """
       Add: self.add_empty()
    """

    def init(self, parent):
        """Finishes initializing the editor by creating the underlying toolkit
        widget.
        """
        # Initialize the trait handler to use:
        trait_handler = self.factory.trait_handler
        if trait_handler is None:
            trait_handler = self.object.base_trait(self.name).handler
        self._trait_handler = trait_handler

        if self.scrollable:
            # Create a scrolled window to hold all of the list item controls:
            self.control = QtGui.QScrollArea()
            self.control.setFrameShape(QtGui.QFrame.Shape.NoFrame)
            self.control.setWidgetResizable(True)
            self._list_pane = QtGui.QWidget()
        else:
            self.control = QtGui.QWidget()
            self._list_pane = self.control
        self._list_pane.setSizePolicy(
            QtGui.QSizePolicy.Policy.Expanding, QtGui.QSizePolicy.Policy.Expanding
        )

        # Create a mapper to identify which icon button requested a contextmenu
        self.mapper = QtCore.QSignalMapper(self.control)

        # Create a widget with a grid layout as the container.
        layout = QtGui.QGridLayout(self._list_pane)
        layout.setAlignment(QtCore.Qt.AlignmentFlag.AlignLeft | QtCore.Qt.AlignmentFlag.AlignTop)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(0)

        # Remember the editor to use for each individual list item:
        editor = self.factory.editor
        if editor is None:
            editor = trait_handler.item_trait.get_editor()
        self._editor = getattr(editor, self.kind)

        # Set up the additional 'list items changed' event handler needed for
        # a list based trait. Note that we want to fire the update_editor_item
        # only when the items in the list change and not when intermediate
        # traits change. Therefore, replace "." by ":" in the extended_name
        # when setting up the listener.
        extended_name = self.extended_name.replace(".", ":")
        self.context_object.on_trait_change(
            self.update_editor_item, extended_name + "_items?", dispatch="ui"
        )
        self.set_tooltip()

    def dispose(self):
        """Disposes of the contents of an editor."""
        self._dispose_items()

        extended_name = self.extended_name.replace(".", ":")
        self.context_object.on_trait_change(
            self.update_editor_item, extended_name + "_items?", remove=True
        )

        super().dispose()

    def update_editor(self):
        """Updates the editor when the object trait changes externally to the
        editor.
        """
        self.mapper = QtCore.QSignalMapper(self.control)
        # Disconnect the editor from any control about to be destroyed:
        self._dispose_items()

        list_pane = self._list_pane
        layout = list_pane.layout()

        # Create all of the list item trait editors:
        trait_handler = self._trait_handler
        resizable = (
            trait_handler.minlen != trait_handler.maxlen
        ) and self.mutable
        item_trait = trait_handler.item_trait

        is_fake = resizable and (len(self.value) == 0)
        if is_fake:
            self.empty_list()
        else:
            self.buttons = []
            # Asking the mapper to send the sender to the callback method
            if is_pyside and QtCore.__version_info__ >= (5, 15):
                self.mapper.mappedInt.connect(self.popup_menu)
            else:
                self.mapper.mapped.connect(self.popup_menu)

        editor = self._editor
        for index, value in enumerate(self.value):
            row, column = divmod(index, self.factory.columns)

            # Account for the fact that we have <columns> number of
            # pairs
            column = column * 2

            if resizable:
                # Connecting the new button to the mapper
                control = IconButton("list_editor.png", self.mapper.map)
                self.buttons.append(control)
                # Setting the mapping and asking it to send the index of the
                # sender to the callback method.  Unfortunately just sending
                # the control does not work for PyQt (tested on 4.11)
                self.mapper.setMapping(control, index)

                layout.addWidget(control, row, column + 1)

            proxy = ListItemProxy(
                self.object, self.name, index, item_trait, value
            )
            if resizable:
                control.proxy = proxy
            peditor = editor(
                self.ui, proxy, "value", self.description, list_pane
            ).trait_set(object_name="")
            peditor.prepare(list_pane)
            pcontrol = peditor.control
            pcontrol.proxy = proxy

            if isinstance(pcontrol, QtGui.QWidget):
                layout.addWidget(pcontrol, row, column)
            else:
                layout.addLayout(pcontrol, row, column)

        # QScrollArea can have problems if the widget being scrolled is set too
        # early (ie. before it contains something).
        if self.scrollable and self.control.widget() is None:
            self.control.setWidget(list_pane)

    def update_editor_item(self, event):
        """Updates the editor when an item in the object trait changes
        externally to the editor.
        """
        # If this is not a simple, single item update, rebuild entire editor:
        if (len(event.removed) != 1) or (len(event.added) != 1):
            self.update_editor()
            return

        # Otherwise, find the proxy for this index and update it with the
        # changed value:
        for control in self._list_pane.children():
            if isinstance(control, QtGui.QLayout):
                continue

            proxy = control.proxy
            if proxy.index == event.index:
                proxy.value = event.added[0]
                break

    def empty_list(self):
        """Creates an empty list entry (so the user can add a new item)."""
        # Connecting the new button to the mapper
        control = IconButton("list_editor.png", self.mapper.map)
        # Setting the mapping and asking it to send the index of the sender to
        # callback method. Unfortunately just sending the control does not
        # work for PyQt (tested on 4.11)
        self.mapper.setMapping(control, 0)
        if is_pyside and QtCore.__version_info__ >= (5, 15):
            self.mapper.mappedInt.connect(self.popup_empty_menu)
        else:
            self.mapper.mapped.connect(self.popup_empty_menu)
        control.is_empty = True
        self._cur_control = control
        self.buttons = [control]

        proxy = ListItemProxy(self.object, self.name, -1, None, None)
        pcontrol = QtGui.QLabel("   (Empty List)")
        pcontrol.proxy = control.proxy = proxy

        layout = self._list_pane.layout()
        layout.addWidget(control, 0, 1)
        layout.addWidget(pcontrol, 0, 0)

    def get_info(self):
        """Returns the associated object list and current item index."""
        proxy = self._cur_control.proxy
        return (proxy.list, proxy.index)

    def popup_empty_menu(self, index):
        """Displays the empty list editor popup menu."""
        self._cur_control = control = self.buttons[index]
        menu = MakeMenu(self.empty_list_menu, self, True, control).menu
        menu.exec_(control.mapToGlobal(QtCore.QPoint(4, 24)))

    def popup_menu(self, index):
        """Displays the list editor popup menu."""
        self._cur_control = sender = self.buttons[index]

        proxy = sender.proxy
        menu = MakeMenu(self.list_menu, self, True, sender).menu
        len_list = len(proxy.list)
        not_full = len_list < self._trait_handler.maxlen

        self._menu_before.enabled(not_full)
        self._menu_after.enabled(not_full)
        self._menu_delete.enabled(len_list > self._trait_handler.minlen)
        self._menu_up.enabled(index > 0)
        self._menu_top.enabled(index > 0)
        self._menu_down.enabled(index < (len_list - 1))
        self._menu_bottom.enabled(index < (len_list - 1))

        menu.exec_(sender.mapToGlobal(QtCore.QPoint(4, 24)))

    def add_item(self, offset):
        """Adds a new value at the specified list index."""
        list, index = self.get_info()
        index += offset
        item_trait = self._trait_handler.item_trait
        if self.factory.item_factory:
            value = self.factory.item_factory(
                *self.factory.item_factory_args,
                **self.factory.item_factory_kwargs,
            )
        else:
            value = item_trait.default_value_for(self.object, self.name)
        try:
            self.value = list[:index] + [value] + list[index:]
        # if the default new item is invalid, we just don't add it to the list.
        # traits will still give an error message, but we don't want to crash
        except TraitError:
            from traitsui.api import raise_to_debug

            raise_to_debug()
        self.update_editor()

    def add_before(self):
        """Inserts a new item before the current item."""
        self.add_item(0)

    def add_after(self):
        """Inserts a new item after the current item."""
        self.add_item(1)

    def add_empty(self):
        """Adds a new item when the list is empty."""
        list, index = self.get_info()
        self.add_item(0)

    def delete_item(self):
        """Delete the current item."""
        list, index = self.get_info()
        self.value = list[:index] + list[index + 1 :]
        self.update_editor()

    def move_up(self):
        """Move the current item up one in the list."""
        list, index = self.get_info()
        self.value = (
            list[: index - 1]
            + [list[index], list[index - 1]]
            + list[index + 1 :]
        )
        self.update_editor()

    def move_down(self):
        """Moves the current item down one in the list."""
        list, index = self.get_info()
        self.value = (
            list[:index] + [list[index + 1], list[index]] + list[index + 2 :]
        )
        self.update_editor()

    def move_top(self):
        """Moves the current item to the top of the list."""
        list, index = self.get_info()
        self.value = [list[index]] + list[:index] + list[index + 1 :]
        self.update_editor()

    def move_bottom(self):
        """Moves the current item to the bottom of the list."""
        list, index = self.get_info()
        self.value = list[:index] + list[index + 1 :] + [list[index]]
        self.update_editor()

    # -- Private Methods ------------------------------------------------------

    def _dispose_items(self):
        """Disposes of each current list item."""
        layout = self._list_pane.layout()
        child = layout.takeAt(0)
        while child is not None:
            control = child.widget()
            if control is not None:
                editor = getattr(control, "_editor", None)
                if editor is not None:
                    editor.dispose()
                    editor.control = None
                control.deleteLater()
            child = layout.takeAt(0)
        del child

    # -- Trait initializers ----------------------------------------------------

    def _kind_default(self):
        """Returns a default value for the 'kind' trait."""
        return self.factory.style + "_editor"

    def _mutable_default(self):
        """Trait handler to set the mutable trait from the factory."""
        return self.factory.mutable

    def _scrollable_default(self):
        return self.factory.scrollable


class CustomEditor(SimpleEditor):
    """Custom style of editor for lists, which displays the items as a series
    of text fields. If the list is editable, an icon next to each item displays
    a menu of operations on the list.
    """

    # -------------------------------------------------------------------------
    #  Class constants:
    # -------------------------------------------------------------------------

    #: Whether the list is displayed in a single row. This value overrides the
    #: default.
    single_row = False


class TextEditor(CustomEditor):

    #: The kind of editor to create for each list item. This value overrides the
    #: default.
    kind = "text_editor"


class ReadonlyEditor(CustomEditor):

    #: Is the list of items being edited mutable? This value overrides the
    #: default.
    mutable = False


class NotebookEditor(Editor):
    """An editor for lists that displays the list as a "notebook" of tabbed
    pages.
    """

    #: The "Close Tab" button.
    close_button = Any()

    #: Maps tab names to QWidgets representing the tab contents
    #: TODO: It would be nice to be able to reuse self._pages for this, but
    #: its keys are not quite what we want.
    _pagewidgets = Dict()

    #: Maps names of tabs to their menu QAction instances; used to toggle
    #: checkboxes
    _action_dict = Dict()

    # -------------------------------------------------------------------------
    #  Trait definitions:
    # -------------------------------------------------------------------------

    #: Is the notebook editor scrollable? This values overrides the default:
    scrollable = True

    #: The currently selected notebook page object:
    selected = Any()

    def init(self, parent):
        """Finishes initializing the editor by creating the underlying toolkit
        widget.
        """
        self._uis = []

        # Create a tab widget to hold each separate object's view:
        self.control = QtGui.QTabWidget()
        self.control.currentChanged.connect(self._tab_activated)

        # minimal dock_style handling
        if self.factory.dock_style == "tab":
            self.control.setDocumentMode(True)
            self.control.tabBar().setDocumentMode(True)
        elif self.factory.dock_style == "vertical":
            self.control.setTabPosition(QtGui.QTabWidget.TabPosition.West)

        # Create the button to close tabs, if necessary:
        if self.factory.deletable:
            button = QtGui.QToolButton()
            button.setAutoRaise(True)
            button.setToolTip("Remove current tab ")
            button.setIcon(ImageResource("closetab").create_icon())

            self.control.setCornerWidget(button, QtCore.Qt.Corner.TopRightCorner)
            button.clicked.connect(self.close_current)
            self.close_button = button

        if self.factory.show_notebook_menu:
            # Create the necessary attributes to manage hiding and revealing of
            # tabs via a context menu
            self._context_menu = QtGui.QMenu()
            self.control.customContextMenuRequested.connect(
                self._context_menu_requested
            )
            self.control.setContextMenuPolicy(QtCore.Qt.ContextMenuPolicy.CustomContextMenu)

        # Set up the additional 'list items changed' event handler needed for
        # a list based trait. Note that we want to fire the update_editor_item
        # only when the items in the list change and not when intermediate
        # traits change. Therefore, replace "." by ":" in the extended_name
        # when setting up the listener.
        extended_name = self.extended_name.replace(".", ":")
        self.context_object.on_trait_change(
            self.update_editor_item, extended_name + "_items?", dispatch="ui"
        )

        # Set of selection synchronization:
        self.sync_value(self.factory.selected, "selected")

    def update_editor(self):
        """Updates the editor when the object trait changes externally to the
        editor.
        """
        # Destroy the views on each current notebook page:
        self.close_all()

        # Create a tab page for each object in the trait's value:
        for object in self.value:
            ui, view_object, monitoring = self._create_page(object)

            # Remember the page for later deletion processing:
            self._uis.append([ui.control, ui, view_object, monitoring])
        self._tab_activated(0)
        if self.selected:
            self._selected_changed(self.selected)

    def update_editor_item(self, event):
        """Handles an update to some subset of the trait's list."""
        index = event.index

        # Delete the page corresponding to each removed item:
        page_name = self.factory.page_name[1:]

        for i in event.removed:
            page, ui, view_object, monitoring = self._uis[index]
            if monitoring:
                view_object.on_trait_change(
                    self.update_page_name, page_name, remove=True
                )
            ui.dispose()
            self.control.removeTab(self.control.indexOf(page))

            if self.factory.show_notebook_menu:
                for name, tmp in self._pagewidgets.items():
                    if tmp is page:
                        del self._pagewidgets[name]
                self._context_menu.removeAction(self._action_dict[name])
                del self._action_dict[name]

            del self._uis[index]

        # Add a page for each added object:
        first_page = None
        for object in event.added:
            ui, view_object, monitoring = self._create_page(object)
            self._uis[index:index] = [
                [ui.control, ui, view_object, monitoring]
            ]
            index += 1

            if first_page is None:
                first_page = ui.control

        if first_page is not None:
            self.control.setCurrentWidget(first_page)

    def close_current(self, force=False):
        """Closes the currently selected tab:"""
        widget = self.control.currentWidget()
        for i in range(len(self._uis)):
            page, ui, _, _ = self._uis[i]
            if page is widget:
                if force or ui.handler.close(ui.info, True):
                    del self.value[i]
                break

        if self.factory.show_notebook_menu:
            # Find the name associated with this widget, so we can purge its action
            # from the menu
            for name, tmp in self._pagewidgets.items():
                if tmp is widget:
                    break
            else:
                # Hmm... couldn't find the widget, assume that we don't need to do
                # anything.
                return

            action = self._action_dict[name]
            self._context_menu.removeAction(action)
            del self._action_dict[name]
            del self._pagewidgets[name]
        return

    def close_all(self):
        """Closes all currently open notebook pages."""
        page_name = self.factory.page_name[1:]

        for _, ui, view_object, monitoring in self._uis:
            if monitoring:
                view_object.on_trait_change(
                    self.update_page_name, page_name, remove=True
                )
            ui.dispose()

        # Reset the list of ui's and dictionary of page name counts:
        self._uis = []
        self._pages = {}

        self.control.clear()

    def dispose(self):
        """Disposes of the contents of an editor."""
        self.context_object.on_trait_change(
            self.update_editor_item, self.name + "_items?", remove=True
        )
        self.close_all()

        super().dispose()

    def update_page_name(self, object, name, old, new):
        """Handles the trait defining a particular page's name being changed."""
        for i, value in enumerate(self._uis):
            page, ui, _, _ = value
            if object is ui.info.object:
                name = None
                handler = getattr(
                    self.ui.handler,
                    "%s_%s_page_name" % (self.object_name, self.name),
                    None,
                )

                if handler is not None:
                    name = handler(self.ui.info, object)

                if name is None:
                    name = str(
                        xgetattr(object, self.factory.page_name[1:], "???")
                    )
                self.control.setTabText(self.control.indexOf(page), name)
                break

    def _create_page(self, object):
        # Create the view for the object:
        view_object = object
        factory = self.factory
        if factory.factory is not None:
            view_object = factory.factory(object)
        ui = view_object.edit_traits(
            parent=self.control, view=factory.view, kind=factory.ui_kind
        ).trait_set(parent=self.ui)

        # Get the name of the page being added to the notebook:
        name = ""
        monitoring = False
        prefix = "%s_%s_page_" % (self.object_name, self.name)
        page_name = factory.page_name
        if page_name[0:1] == ".":
            name = xgetattr(view_object, page_name[1:], None)
            monitoring = name is not None
            if monitoring:
                handler_name = None
                method = getattr(self.ui.handler, prefix + "name", None)
                if method is not None:
                    handler_name = method(self.ui.info, object)
                if handler_name is not None:
                    name = handler_name
                else:
                    name = str(name) or "???"
                view_object.on_trait_change(
                    self.update_page_name, page_name[1:], dispatch="ui"
                )
            else:
                name = ""
        elif page_name != "":
            name = page_name

        if name == "":
            name = user_name_for(view_object.__class__.__name__)

        # Make sure the name is not a duplicate:
        if not monitoring:
            self._pages[name] = count = self._pages.get(name, 0) + 1
            if count > 1:
                name += " %d" % count

        # Return the control for the ui, and whether or not its name is being
        # monitored:
        image = None
        method = getattr(self.ui.handler, prefix + "image", None)
        if method is not None:
            image = method(self.ui.info, object)

        if image is None:
            self.control.addTab(ui.control, name)
        else:
            self.control.addTab(ui.control, image, name)

        if self.factory.show_notebook_menu:
            newaction = self._context_menu.addAction(name)
            newaction.setText(name)
            newaction.setCheckable(True)
            newaction.setChecked(True)
            newaction.triggered.connect(
                lambda e, name=name: self._menu_action(e, name=name)
            )
            self._action_dict[name] = newaction
            self._pagewidgets[name] = ui.control

        return (ui, view_object, monitoring)

    def _tab_activated(self, idx):
        """Handles a notebook tab being "activated" (i.e. clicked on) by the
        user.
        """
        widget = self.control.widget(idx)
        for page, ui, _, _ in self._uis:
            if page is widget:
                self.selected = ui.info.object
                break

    def _selected_changed(self, selected):
        """Handles the **selected** trait being changed."""
        for page, ui, _, _ in self._uis:
            if ui.info and selected is ui.info.object:
                self.control.setCurrentWidget(page)
                break
            deletable = self.factory.deletable
            deletable_trait = self.factory.deletable_trait
            if deletable and deletable_trait:
                enabled = xgetattr(selected, deletable_trait, True)
                self.close_button.setEnabled(enabled)

    def _context_menu_requested(self, event):
        self._context_menu.popup(self.control.mapToGlobal(event))

    def _menu_action(self, event, name=""):
        """Qt signal handler for when a item in a context menu is actually
        selected.  Not that we get this even after the underlying value has
        already changed.
        """
        action = self._action_dict[name]
        checked = action.isChecked()
        if not checked:
            for ndx in range(self.control.count()):
                if self.control.tabText(ndx) == name:
                    self.control.removeTab(ndx)
        else:
            # TODO: Fix tab order based on the context_object's list
            self.control.addTab(self._pagewidgets[name], name)



# (C) Copyright 2004-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" Tests pertaining to the ImageEditor
"""

import unittest

import pkg_resources

from pyface.api import Image, ImageResource
from traits.api import File, HasTraits
from traitsui.api import ImageEditor, Item, View
from traitsui.tests._tools import (
    BaseTestMixin,
    create_ui,
    is_qt,
    is_mac_os,
    requires_toolkit,
    ToolkitName,
)


filename1 = pkg_resources.resource_filename(
    "traitsui", "examples/demo/Extras/images/python-logo.png"
)
filename2 = pkg_resources.resource_filename(
    "traitsui", "examples/demo/Extras/images/info.png"
)


class ImageDisplay(HasTraits):

    image = Image()


@requires_toolkit([ToolkitName.wx, ToolkitName.qt])
class TestImageEditor(BaseTestMixin, unittest.TestCase):

    def test_image_editor_static(self):
        obj1 = ImageDisplay()
        view = View(
            Item(
                'image',
                editor=ImageEditor(
                    image=ImageResource(filename1),
                ),
            )
        )

        # This should not fail.
        with create_ui(obj1, dict(view=view)) as ui:
            pass

    def test_image_editor_resource(self):
        obj1 = ImageDisplay(
            image=ImageResource(filename1)
        )
        view = View(
            Item(
                'image',
                editor=ImageEditor()
            )
        )

        # This should not fail.
        with create_ui(obj1, dict(view=view)) as ui:
            obj1.image = ImageResource(filename2)

    def test_image_editor_array(self):
        try:
            import numpy as np
            from pyface.api import ArrayImage
        except ImportError:
            self.skipTest("NumPy is not available")

        gradient1 = np.empty(shape=(256, 256, 3), dtype='uint8')
        gradient1[:, :, 0] = np.arange(256).reshape(256, 1)
        gradient1[:, :, 1] = np.arange(256).reshape(1, 256)
        gradient1[:, :, 2] = np.arange(255, -1, -1).reshape(1, 256)

        gradient2 = np.empty(shape=(256, 256, 3), dtype='uint8')
        gradient2[:, :, 0] = np.arange(255, -1, -1).reshape(256, 1)
        gradient2[:, :, 1] = np.arange(256).reshape(1, 256)
        gradient2[:, :, 2] = np.arange(255, -1, -1).reshape(1, 256)

        obj1 = ImageDisplay(
            image=ArrayImage(data=gradient1)
        )
        view = View(
            Item(
                'image',
                editor=ImageEditor()
            )
        )

        # This should not fail.
        with create_ui(obj1, dict(view=view)) as ui:
            obj1.image = ArrayImage(data=gradient2)

    @unittest.skipIf(is_mac_os, "Segfault on MacOS, see issue #1979")
    def test_image_editor_pillow(self):
        try:
            import PIL.Image
            from pyface.api import PILImage
        except ImportError:
            self.skipTest("Pillow is not available")
        if is_qt:
            try:
                # is ImageQt available as well
                from PIL.ImageQt import ImageQt
            except ImportError:
                self.skipTest("ImageQt is not available")

        pil_image_1 = PIL.Image.open(filename1)
        pil_image_2 = PIL.Image.open(filename2)

        obj1 = ImageDisplay(
            image=PILImage(image=pil_image_1)
        )
        view = View(
            Item(
                'image',
                editor=ImageEditor()
            )
        )

        # This should not fail.
        with create_ui(obj1, dict(view=view)) as ui:
            obj1.image = PILImage(image=pil_image_2)

    def test_image_editor_none(self):

        obj1 = ImageDisplay(image=None)
        view = View(
            Item(
                'image',
                editor=ImageEditor()
            )
        )

        # This should not fail.
        with create_ui(obj1, dict(view=view)) as ui:
            pass



# (C) Copyright 2004-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" Defines the various text editors for the wxPython user interface toolkit.
"""


import wx

from traits.api import TraitError

from traitsui.editors.text_editor import evaluate_trait

from .editor import Editor

from .editor_factory import ReadonlyEditor as BaseReadonlyEditor

from .constants import OKColor


# Readonly text editor with view state colors:
HoverColor = wx.LIGHT_GREY
DownColor = wx.WHITE


class SimpleEditor(Editor):
    """Simple style text editor, which displays a text field."""

    #: Flag for window styles:
    base_style = 0

    #: Background color when input is OK:
    ok_color = OKColor

    # -------------------------------------------------------------------------
    #  Trait definitions:
    # -------------------------------------------------------------------------

    #: Function used to evaluate textual user input:
    evaluate = evaluate_trait

    def init(self, parent):
        """Finishes initializing the editor by creating the underlying toolkit
        widget.
        """
        factory = self.factory
        style = self.base_style
        self.evaluate = factory.evaluate
        self.sync_value(factory.evaluate_name, "evaluate", "from")

        if (not factory.multi_line) or factory.password:
            style &= ~wx.TE_MULTILINE

        if factory.password:
            style |= wx.TE_PASSWORD

        multi_line = (style & wx.TE_MULTILINE) != 0
        if multi_line:
            self.scrollable = True

        if factory.enter_set and (not multi_line):
            control = wx.TextCtrl(
                parent, -1, self.str_value, style=style | wx.TE_PROCESS_ENTER
            )
            parent.Bind(
                wx.EVT_TEXT_ENTER, self.update_object, id=control.GetId()
            )
        else:
            control = wx.TextCtrl(parent, -1, self.str_value, style=style)

        control.Bind(wx.EVT_KILL_FOCUS, self.update_object)
        if control.IsSingleLine():
            control.SetHint(self.factory.placeholder)

        if factory.auto_set:
            parent.Bind(wx.EVT_TEXT, self.update_object, id=control.GetId())

        self.control = control
        self.set_error_state(False)
        self.set_tooltip()

    def update_object(self, event):
        """Handles the user entering input data in the edit control."""
        if isinstance(event, wx.FocusEvent):
            # Ensure that the base class' focus event handlers are run, some
            # built-in behavior may break on some platforms otherwise.
            event.Skip()

        if (not self._no_update) and (self.control is not None):
            try:
                self.value = self._get_user_value()

                if self._error is not None:
                    self._error = None
                    self.ui.errors -= 1

                self.set_error_state(False)

            except TraitError as excp:
                pass

    def update_editor(self):
        """Updates the editor when the object trait changes externally to the
        editor.
        """
        user_value = self._get_user_value()
        try:
            unequal = bool(user_value != self.value)
        except ValueError:
            # This might be a numpy array.
            unequal = True

        if unequal:
            self._no_update = True
            self.control.SetValue(self.str_value)
            self._no_update = False

        if self._error is not None:
            self._error = None
            self.ui.errors -= 1
            self.set_error_state(False)

    def _get_user_value(self):
        """Gets the actual value corresponding to what the user typed."""
        value = self.control.GetValue()
        try:
            value = self.evaluate(value)
        except:
            pass

        try:
            ret = self.factory.mapping.get(value, value)
        except TypeError:
            # The value is probably not hashable:
            ret = value

        return ret

    def error(self, excp):
        """Handles an error that occurs while setting the object's trait value."""
        if self._error is None:
            self._error = True
            self.ui.errors += 1

        self.set_error_state(True)

    def in_error_state(self):
        """Returns whether or not the editor is in an error state."""
        return self.invalid or self._error


class CustomEditor(SimpleEditor):
    """Custom style of text editor, which displays a multi-line text field."""

    #: Flag for window style. This value overrides the default.
    base_style = wx.TE_MULTILINE


class ReadonlyEditor(BaseReadonlyEditor):
    """Read-only style of text editor, which displays a read-only text field."""

    def init(self, parent):
        """Finishes initializing the editor by creating the underlying toolkit
        widget.
        """
        super().init(parent)

        if self.factory.view is not None:
            control = self.control
            control.Bind(wx.EVT_ENTER_WINDOW, self._enter_window)
            control.Bind(wx.EVT_LEAVE_WINDOW, self._leave_window)
            control.Bind(wx.EVT_LEFT_DOWN, self._left_down)
            control.Bind(wx.EVT_LEFT_UP, self._left_up)

    def update_editor(self):
        """Updates the editor when the object trait changes externally to the
        editor.
        """
        control = self.control
        new_value = self.str_value

        if hasattr(self.factory, "password") and self.factory.password:
            new_value = "*" * len(new_value)

        if (self.item.resizable is True) or (self.item.height != -1.0):
            if control.GetValue() != new_value:
                control.SetValue(new_value)
                control.SetInsertionPointEnd()

        elif control.GetLabel() != new_value:
            control.SetLabel(new_value)

    def dispose(self):
        """Disposes of the contents of an editor."""
        if self.factory.view is not None:
            control = self.control
            control.Unbind(wx.EVT_ENTER_WINDOW)
            control.Unbind(wx.EVT_LEAVE_WINDOW)
            control.Unbind(wx.EVT_LEFT_DOWN)
            control.Unbind(wx.EVT_LEFT_UP)

        super().dispose()

    # -- Private Methods ------------------------------------------------------

    def _set_color(self):
        control = self.control
        if not self._in_window:
            color = control.GetParent().GetBackgroundColour()
        elif self._down:
            color = DownColor
        else:
            color = HoverColor

        control.SetBackgroundColour(color)
        control.Refresh()

    # -- wxPython Event Handlers ----------------------------------------------

    def _enter_window(self, event):
        self._in_window = True
        self._set_color()

    def _leave_window(self, event):
        self._in_window = False
        self._set_color()

    def _left_down(self, event):
        self.control.CaptureMouse()
        self._down = True
        self._set_color()

    def _left_up(self, event):
        self._set_color()
        if not self._down:
            return

        self.control.ReleaseMouse()
        self._down = False

        if self._in_window:
            self.object.edit_traits(
                view=self.factory.view, parent=self.control
            )


TextEditor = SimpleEditor



# (C) Copyright 2004-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

"""
Control the height and width of a Group

TraitsUI does not permit explicit specification of the height or width of a
Group (or its descendants). The workaround is to put each Group whose size you
wish to control into its own View class, which can be an Item (hence can be
size-controlled) in the larger View. Sometimes it is necessary to repeat such
surgery at several levels to get the desired layout.

We separate the left and right groups by a splitter (HSplit), and also
make the window itself resizable.

"""
from numpy import linspace, pi, sin
from traits.api import Code, HasTraits, Instance, Str, Int

# UItem is Unlabeled Item
from traitsui.api import (
    View,
    Item,
    UItem,
    HSplit,
    InstanceEditor,
    VGroup,
    HGroup,
)


class InstanceUItem(UItem):
    """Convenience class for including an Instance in a View"""

    style = Str('custom')
    editor = Instance(InstanceEditor, ())


class CodeView(HasTraits):
    """Defines a sub-view whose size we wish to explicitly control."""

    n = Int(123)
    code = Code()
    view = View(
        # This HGroup keeps 'n' from over-widening, by implicitly placing
        # a spring to the right of the item.
        HGroup(Item('n')),
        UItem('code'),
        resizable=True,
    )


class VerticalBar(HasTraits):
    """Defines a sub-view whose size we wish to explicitly control."""

    a = Str('abcdefg')
    b = Int(123)
    view = View(
        VGroup(
            Item('a'),
            Item('b'),
            show_border=True,
        ),
    )


class BigView(HasTraits):
    """Defines the top-level view. It contains two side-by-side panels (a
    vertical bar and a code editor under an integer) whose relative sizes we
    wish to control explicitly. If these were simply defined as Groups, we
    could not control their sizes. But by extracting them as separate classes
    with their own views, the resizing becomes possible, because they are
    loaded as Items now.
    """

    bar = Instance(VerticalBar, ())
    code = Instance(CodeView)
    view = View(
        HSplit(
            # Specify pixel width of each sub-view. (Or, to specify
            # proportionate width, use a width < 1.0)
            # We specify the height here for sake of demonstration;
            # it could also be specified for the top-level view.
            InstanceUItem('bar', width=150),
            InstanceUItem('code', width=500, height=500),
            show_border=True,
        ),
        resizable=True,
    )


x = linspace(-2 * pi, 2 * pi, 100)
pv = CodeView(code=open(__file__).read())
bv = BigView(code=pv)
bv.configure_traits()
